#define BIG_NUMBER 2000000002
#include "robot.h"

typedef struct
{
	void *next;
	int x,y,z;
} dot_t;

goal_t *evaluate(goal_t *goal_in);
void do_move(char *move, goal_t *goal);

void free_path(path_t **path_in)
{
	path_t *path=*path_in;
	if (path)
	{
		if (path->ds)
		{
			int nx;
			for (nx=0; nx <= world_x; nx++)
				free(path->ds[nx]);
			free(path->ds);
		}
		free (path);
	}
	*path_in=NULL;
}

void free_goal(goal_t **goal_in)
{
	goal_t *goal=*goal_in;

	*goal_in = NULL;
	if (goal)
	{
		if (goal->path)
		{
			free_path(&goal->path);
		}
		free (goal);
	}
	*goal_in=NULL;
}

#if 0  // virtual path maybe useful.  maybe not.
				int robot_factor;
				robot_factor=10;
				while (x!=pkg->x || y!=pkg->y)
				{
					robot_t *robot;
					char c;
					int i, ldist=BIG_NUMBER;

					/* virtual move */
					for (dir=s_nsew; *dir; dir++)
						if (pkg->path->ds[x+DX(*dir)][y+DY(*dir)]<ldist)
						{
							ldist=pkg->path->ds[x+DX(*dir)][y+DY(*dir)];
							c=*dir;
						}
					x+=DX(c); y+=DY(c);
					
				} // done calculating distance
#endif

/*
	if (   (rc=robot_proximity(x,y,robot_factor)) 
	    && robot_factor<10)
		robot_factor++;
	zzz += rc/5;
*/

int robot_proximity(x, y, weight)
{
	int count=0;
	int i;
	robot_t *robot;

	/* look around */
	for (i=0; i<PROX; i++)
		for (robot=robots; robot; robot=robot->next)
	{
		int px = x+proximity[i][0]; 
		int py = y+proximity[i][1];
		if ( robot->x==px && robot->y==py )
		{
			count+=weight*proximity[i][2];
		}
	}
	return count;
}

path_t *find_path(int x, int y, int flag_water)
{
//	int robot_factor=rand()%3;
	int robot_factor=2;
	int factor=0;
	int     nx, ny;
	char   *dir;
	dot_t  *newdot, *dotlist, *dot;
	path_t *path;

	if (x<=0 || y<=0 || x> world_x || y>world_y)
		return NULL;

	if (n_robots > 0)  factor=1;
	if (n_robots > 3)  factor=2;
	if (n_robots > 7)  factor=4;
	if (n_robots > 15) factor=8;

	/*if (n_robots>0)*/ //factor=13; // testing..
	
	path = malloc(sizeof(path_t));
	path->n_robots = n_robots;

	/* init */
	dotlist = malloc(sizeof(dot_t));
	dotlist->next=NULL ; dotlist->x=x ; dotlist->y=y ; dotlist->z=1;

	path->ds=malloc(sizeof(int *) * world_x+2);
	for (nx=0; nx <= world_x; nx++)
		path->ds[nx]=malloc(sizeof(int)*(world_y+2));
	for(nx=0;nx<=world_x;nx++)for(ny=0;ny<=world_y;ny++) path->ds[nx][ny]=0;

	path->ds[x][y] = dotlist->z = 1;

	//xxx should check for overflow down there somewhere, in case of large world
	
	/* run */
	while ( dotlist )
	{
		for ( dir = s_nsew; *dir; dir++ )
		{
			nx = (dotlist->x)+DX(*dir); ny = (dotlist->y)+DY(*dir);
			if ( (world[nx][ny].passable) && (path->ds[nx][ny]==0) )
			{
				/* add new one */
				newdot    = malloc(sizeof(dot_t));
				if (newdot==NULL) fprintf(stderr, "foobar.\n");
				newdot->x = nx; 
				newdot->y = ny; 
				newdot->z = dotlist->z+10;

				//stay clear of water, walls, and maybe even robots that don't move.
				if (flag_water)
				{
					int wallcount=0;
					int walls=0;
					int water=0;
					int i, px, py;

					for (i=0; i<PROX; i++)
					{
						px = nx+proximity[i][0]; 
						py = ny+proximity[i][1];

						if (px>=0 && py>=0 && px<=world_x+1 && py<=world_y+1)
						{
							if (world[px][py].type==WATER)
							{
								water+=proximity[i][2]*(factor+1);
								if (1 == (  proximity[i][0]*proximity[i][0] 
								           + proximity[i][1]*proximity[i][1] ))
									wallcount++;
							}
							if (world[px][py].type==WALL*factor)
							{
								walls+=proximity[i][2];

								if (1 == (  proximity[i][0]*proximity[i][0] 
								           + proximity[i][1]*proximity[i][1] ))
									wallcount++;
							}

						}
					}
					newdot->z += (water/3 + walls/7);
					if (wallcount>1)
						newdot->z += 3*factor;
				}
				if  (flag_water==3 
				 || (flag_water==4 && ((DIST(me.x,nx) + DIST(me.y,ny))<4)) )
				{
					int rc;
					/* look around for robots */
					if (   (rc=robot_proximity(nx,ny,robot_factor)) 
					    && robot_factor<flag_water)
						robot_factor++;
					newdot->z += rc;

					// needs more than normal influence here to work.
					// need to balance... xxx
					if (  3 > ((nx-me.x)*(nx-me.x) + (ny-me.y)*(ny-me.y)) )
						newdot->z += rc;
				}

				/* add to list - lowest z first */
				for (dot=dotlist; dot; dot=dot->next)
				{
					if ((!(dot->next)) || ((dot_t*)dot->next)->z > newdot->z)
					{
						newdot->next = dot->next;
						dot->next = newdot;
						break;
					}
				}
				if (!dot) printf("SKSKSKSKS\n");
				path->ds[nx][ny]=newdot->z;
			} //end if(newdot)
		} //end for(direction)

		/* remove the one we've dealt with */
		dot = dotlist->next; free(dotlist); dotlist=dot;
	}

#	if 0
	printf ("Path to (%d,%d):\n", x, y);
	for (y=0; y<world_x; y++)
	{
		for (x=0; x<world_x; x++)
			printf("% 2d", path->ds[x][y]/100);
		printf("\n");
	}
#	endif
	return path;
}

int water_threats(char *move)
{
	robot_t *robot;
	int nearby;

	char threat_dir, move_dir='\0';
	char *dir;
	char c_way[2];
	int  v_way[2];

	nearby = robot_proximity(me.x,me.y,1);

	/* check for threat to me */
	threat_dir = '\0';
	for (dir=s_nsew; *dir; dir++)
	{
		int nx = me.x+DX(*dir);
		int ny = me.y+DY(*dir);
		if (world[nx][ny].passable)
		{
			for (robot=robots; robot; robot=robot->next)
				if (robot->x==nx && robot->y==ny)
					break;
			if (robot && world[me.x-DX(*dir)][me.y-DY(*dir)].type==WATER)
			{
				threat_dir = *dir;
				break;
			}
		}
	}
	if (threat_dir)
	{
		int bid, i;
		int flag_robot=0;

		move_dir='\0';
		v_way[0] = v_way[1] = 0;
		if (threat_dir=='E' || threat_dir=='W') { c_way[0]='N'; c_way[1]='S'; }
		else                                    { c_way[0]='E'; c_way[1]='W'; }

		/* which way to run... */
		for (i=0; i<2; i++)
		{
			int nx = me.x+DX(*dir);
			int ny = me.y+DY(*dir);
			if (!(world[me.x+DX(c_way[i])][me.x+DY(c_way[i])].passable))
			{
				move_dir = c_way[i==0?1:0];
				break;
			}
			if (!main_goal) main_goal = evaluate(main_goal);
			for (robot = robots; robot; robot=robot->next)
				if (robot->x==nx && robot->y==ny)
					flag_robot=1;

			v_way[i]=main_goal->path->ds[nx][ny] + (robot ? 10 : 0) + (rand()%5);
		}
		if (!move_dir) move_dir = v_way[0] < v_way[1] ? c_way[0] : c_way[1];

		/* make a move... */
		/* calculate a bid */
		bid = nearby;
		me.angry+=6;
		bid += 4 * (flag_robot ? 2 : 0) * me.angry;
		if (me.money>10 && bid>me.money/10) bid = me.money/10;
		if (bid<=0) bid = 1;

		me.money-= (bid>0?bid:-bid);
		sprintf(move, "%d Move %c\n", bid, move_dir);
		lastmove = move_dir;
		return 1;
	}  //end threat to me.

	/* check for threat to others */
	threat_dir = '\0';
	for (dir=s_nsew; *dir; dir++)
	{
		int nx = me.x+DX(*dir);
		int ny = me.y+DY(*dir);
		if (world[nx][ny].passable)
		{
			for (robot=robots; robot; robot=robot->next)
				if (robot->x==nx && robot->y==ny)
					break;
			if (robot && world[nx+DX(*dir)][ny+DY(*dir)].type==WATER)
			{
				me.angry += 8;  /* just in case he notices too, and gets weird. */
				threat_dir = *dir;
				break;
			}
		}
	}
	if (threat_dir && nearby<=10 && (!me.packages||me.angry>33)) 
	{
		// give him a gentle nudge.
		int bid = 2+(rand()%(me.angry+2));
		if (me.money>26 && bid>me.money/13) bid = me.money/13;
		if (bid<=2) bid = 2;
		if (me.money==1) bid = 1;

		me.money-= (bid>0?bid:-bid);
		sprintf(move, "%d Move %c\n", bid, threat_dir);
		lastmove = move_dir;
		return 1;
	}
	return 0;
}

//XXX NEED TO check for easy pickups nearby, even if going elsewhere
goal_t *evaluate(goal_t *goal_in)
{
	int value=-BIG_NUMBER;
	int val;
	goal_t *goal=goal_in;
	goal_t *goal_tmp=NULL;
	package_t *pkg;
	base_t *base;
	base_t *selected_base=NULL;

	free_goal(&goal);
	
	if (!goal)
	{
		value = val = -BIG_NUMBER;
		for (base = bases ; base ; base=base->next)
		{
			int rp=robot_proximity(base->x, base->y, 1);

			goal_tmp = malloc(sizeof(goal_t));
			memset(goal_tmp, 0x00, sizeof(goal_t));
			goal_tmp->type = HOME;
			goal_tmp->x=base->x; goal_tmp->y=base->y;

			if (!(goal_tmp->path = find_path(base->x, base->y, 4)))
			{
				printf("Hmm... an unreachable base.\n");
			} else
			{
				int n_pkgs=0;
				val=0-goal_tmp->path->ds[me.x][me.y];

				for (pkg=world[base->x][base->y].packages;pkg;pkg=pkg->wnext)
					if (pkg->weight <= me.max_load-me.load)
						n_pkgs++;

				if (base->visited && n_pkgs==0)
				{
					val -= 10000;
					val *= 10;

					if (   turn-(base->visited) < 200 )
						val -= BIG_NUMBER/2;
				}
				else
					val += 8*(n_pkgs>20?20:n_pkgs); 

#				if 0  //not worth it, I think.
					if ( base->visited && n_pkgs < 3 && turn-base->visited < 1000 )
						val *= 2 * (300 - (turn-base->visited));

					if ( base->visited && n_pkgs < 1 && turn-base->visited < 1000 )
						val *= 3 * (300 - (turn-base->visited));

					if ( base->visited && n_pkgs==0 && turn-base->visited < 1000 )
						val=-100000;
#				endif

				val+=rand()%2;
				val-=rp;
				if (val > value)
				{
					selected_base = base;
					value = val;
					free_goal(&goal);
					goal = goal_tmp;
					goal_tmp=NULL;
				}
			}
			free_goal(&goal_tmp);
		}

#if 1 //{
		// check for stray packages
		for (pkg=packages; pkg; pkg=pkg->next)
			if (   pkg->owner==NULL 
			    && pkg->x>0 && pkg->y>0 && world[pkg->x][pkg->y].type!=BASE
				&& pkg->weight < me.max_load-me.load)
		{
			int rp=robot_proximity(pkg->x, pkg->y, 1);

			goal_tmp = malloc(sizeof(goal_t));
			memset(goal_tmp, 0x00, sizeof(goal_t));
			goal_tmp->x=pkg->x; goal_tmp->y=pkg->y;
			goal_tmp->type = PICKUP;
			if (!(goal_tmp->path = find_path(pkg->x, pkg->y, 4)))
			{
				printf("My omniscient brain senses an unreachable package.\n");
			} else
			{
				package_t *p_pkg;
				int n_pkgs=0;
				int flag_dest=0;

				val=0 - (goal_tmp->path->ds[me.x][me.y]*2);

				for (p_pkg=world[pkg->x][pkg->y].packages;p_pkg;p_pkg=p_pkg->wnext)
					if (p_pkg->weight <= me.max_load-me.load)
				{
					n_pkgs++;
					if (p_pkg->dest_x>0 && p_pkg->dest_y>0)
						flag_dest=1;
				}

				if (n_pkgs<3 && !flag_dest)   //all unknown
					val *= 10;   // only if we've got nothing else to do.
				else if (n_pkgs<2)
					val *= 3;    // 3x bases?  hmm...
				else
					val *= 2;    //xxx need to tune this

				val-=(rand()%me.angry);
				val-=rp;
				if (val > value)
				{
					value = val;
					free_goal(&goal);
					goal = goal_tmp;
					goal_tmp=NULL;
					selected_base = NULL;
				}
			}
			free_goal(&goal_tmp);
		}	
#endif //}

		//package in hand is worth n in the bush //xxx tune this
		if (goal && value> -BIG_NUMBER/16) 
		{
			package_t *p_pkg;
			value *= 2;
			if (robot_proximity(goal->x, goal->y, 1))
			{
				int n_pkgs=0;
				value *= 2;   //not if robots near there, either

				for (p_pkg=world[goal->x][goal->y].packages;p_pkg;p_pkg=p_pkg->wnext)
					if (p_pkg->weight==0 || p_pkg->weight <= me.max_load-me.load)
					n_pkgs++;

				if (n_pkgs<3)
					value *= 2;

				if (n_pkgs==0)
					value = -BIG_NUMBER / 10;
			}
			if (robot_proximity(me.x, me.y, 1)>7 || me.angry>10)
				value *= 2;
			if (me.angry>20)
			{
				value = -BIG_NUMBER;  //too angry to bother.
			}
		}

		//Check packages in hand
		for (pkg=me.packages; pkg; pkg=pkg->mnext)
		{
			if (pkg->dest_x==me.x&&pkg->dest_y==me.y)
				continue;

			goal_tmp = malloc(sizeof(goal_t));
			memset(goal_tmp, 0x00, sizeof(goal_t));
			goal_tmp->x=pkg->dest_x; goal_tmp->y=pkg->dest_y;
			goal_tmp->type = DELIVERY;

			if (!(goal_tmp->path = find_path(pkg->dest_x, pkg->dest_y, 3)))
			{
				printf("This is bad: I've got an undeliverable package.\n");
			} else
			{
				// just pick the closest one.
				val=0-goal_tmp->path->ds[me.x][me.y];
				if (val > value)
				{
					value = val;
					free_goal(&goal);
					goal = goal_tmp;
					goal_tmp=NULL;
				}
			}
			free_goal(&goal_tmp);
		}
		if (goal)    //deliveries come first, for now...xxx
		{
			// Made the choice, now go for it straight.
			free_path(&goal->path);
			goal->path=find_path(goal->x, goal->y, 4);
			return goal;
		}

		if (!goal)
		{
			printf("huh? An empty and desolate place.\n");
			goal = malloc(sizeof(goal_t));
			memset(goal, 0x00, sizeof(goal_t));
			goal->x=me.x; goal->y=me.y;
			goal->type = DELIVERY;
			goal->path = find_path(me.x, me.y, 1);
			return goal;
		}
	}
	return goal;
}

// no time to solve the travelling-salesman analogue.
// well, I'll hack it, heuristically-like.
char picklist[8192];
int do_pickups(char *move)
{
	char      *dir;
	int        new_distance;
	int        total_distance=0;
	int        weight=0;
	int        near_robots=0;
	package_t *pkg, *last_pkg=NULL;
	char      *move_p=picklist;
	int        bid          = 1;
	int        flag         = 1; //pickup flag
	int        me_count     = 0; //current count
	int        pick_count   = 0; //count of picks

	for (pkg=world[me.x][me.y].packages; pkg; pkg=pkg->wnext)
		pkg->pickup_flag=0;

	// Consider the last of the packages we're already holding.
	for (pkg=me.packages; pkg; pkg=pkg->mnext)
	{
		last_pkg = pkg;
		me_count++;
	}

	if (last_pkg && me.angry > 500)
	{
		// If I'm real angry, make do with packages I have.
		return 0;
	}

	// Find the "closest" sequence off the closest available pkg.
	// I goes for distance, rather than weight.
	while(flag)  // Keep picking until done.
	{
		int distance;
		int count=0; 
		package_t *pick=NULL;

		distance=BIG_NUMBER;
		flag=0;

		for (pkg=world[me.x][me.y].packages; pkg; pkg=pkg->wnext)
		{
			if (   !pkg->pickup_flag
				&& pkg->weight <= me.max_load - me.load - weight
				&& pkg->weight && pkg->dest_x>0 && pkg->dest_y>0 )
			{
				int x,y;
				x = (last_pkg ? last_pkg->dest_x : me.x);
				y = (last_pkg ? last_pkg->dest_y : me.y);

				/* For the moment at least, the worldlist/path is destroyed
				 * on package update, so they'll all get recreated here. 
				 * Which is fine, since they can reflect current robot state.
				 */
				if (!pkg->path && !(pkg->path=find_path(pkg->dest_x, pkg->dest_y, 3)))
				{
					fprintf(stderr, "\n\nNO PATH\b");fflush(stderr);
					continue;
				}
				new_distance = pkg->path->ds[x][y];

				if (   new_distance<distance 
					|| (new_distance==distance && (rand()%2==0)) )
				{
					/* pick this one! */
					total_distance += new_distance;
					distance        = new_distance;
					pick=pkg;
				}
				count++;
			} // done evaluating this item
		} // end item loop 
		
		// pick up the best one.
		if (pick)
		{
			flag = 1;
			pick_count++;
			pick->pickup_flag = 1;
			last_pkg          = pick;
			move_p           += sprintf(move_p, " %d", pick->n);
			weight           += pick->weight;
		}
	} // picklist loop

	if (move_p==picklist) return 0;  // All that work for nothin'

	/* now to calculate the bid ... */
	/* check for nearby competitors */
	near_robots=0;
	for (dir=s_nsew; *dir; dir++)
	{
		int nx, ny;
		robot_t *robot;

		nx=me.x+DX(*dir); ny=me.y+DY(*dir);

		for (robot=robots; robot; robot=robot->next)
			if (robot->x==nx && robot->y==ny)
			{
				bid*=2; near_robots++;

				if (world[me.x-DX(*dir)][me.y-DY(*dir)].type==WATER)
				{
					// gonna need a somewhat large bid methinks.
					bid += 13;
					bid += robot->angry + me.angry / 4;
					// better to move instead?
				}
			}
	}

	/* Now we just calculate what would be left if we allow other nearby robots 
	 * (those that look likely to do so) to pick from this pile, and then we 
	 * adjust the bid accordingly.  Or, not.
	 */

	// see I've got better things to do.
	if (me.angry>10 && near_robots && me_count >= pick_count)
	{
		// XXX do something here!
		sprintf(move, "-1 Drop\n");
		return 0;
	}
	if (near_robots && me_count+pick_count==1)
	{
		if (rand()%me.angry > 50)
		{
			// XXX do something here!
			me.money--;
			sprintf(move, "-1 Drop\n");
			return 0;
		}
	}

	/* How much is all this worth, then? */
	if (robot_proximity(me.x, me.y, 1)==0) 
		bid = 1;
	else
	{
		bid += (near_robots*3 + me.angry) / 5;
		if (!me.packages)      bid /= 4;   //
		if (bid>(me.money/10)) bid  = me.money/10;
		if (bid==0)            bid  = 1;
//		if (pick_count<=1)     bid  = 1; else bid+=me_count+pick_count;
	}
	me.money-= (bid>0?bid:-bid);
	sprintf(move, "%d Pick%s\n", bid, picklist);
	lastmove='P';
	if (bid>4)
	{
		free_goal(&main_goal);
	}

	return 1;
}

int drop_packages(char *move)
{
	package_t *pkg;
	char *move_p;
	int bid=2;
	int flag=0;
	int count=0;

	move_p=move;

	for (pkg=me.packages; pkg; pkg=pkg->mnext)
		count++;

	if (count==1 && robot_proximity(me.x,me.y)) bid=-2;

	for (pkg=me.packages; pkg; pkg=pkg->mnext)
	{
		if (pkg->dest_x==me.x && pkg->dest_y==me.y)
		{
			if (!flag)
			{
				move_p+=sprintf(move_p, "%d Drop", bid);
				lastmove='D';
			}
			flag++;
			move_p+=sprintf(move_p, " %d", pkg->n);
		}
	}
	if (flag) 
	{
		move_p+=sprintf(move_p, "\n");
		me.money-= (bid>0?bid:-bid);
	}
	return flag;
}

void do_move(char *move, goal_t *goal)
{
	robot_t *robot;
	int near_robots=0, total_near_robots=0, flag_near = 0;
	char   *dir, move_dir;
	int     bid=1;
	int i, nx, ny;
	int flag_water=0;
	int move_obstruction;

	int randomfactor;

	if (!goal || !goal->path)
	{
		sprintf(gmessage, "I'm stuck!\n");
		fprintf(stderr, "\n\n\nI'm stuck!\n");
		sprintf(move, "-1 Drop\n");
		me.money--;
		return;
	}

	// just in case....
	if (goal->x==me.x && goal->y==me.y)
	{
		free_goal(&main_goal);
		goal = main_goal = evaluate(main_goal);
	}
	if (goal->x==me.x && goal->y==me.y)
	{
		int rp=20000;
		move_dir=s_nsew[rand()%4];
		printf("Bad move, dude!\n");
		for (dir=s_nsew; *dir; dir++)
		{
			if ( world[me.x+DX(*dir)][me.y+DY(*dir)].passable 
				&& robot_proximity(me.x+DX(*dir), me.y+DY(*dir), 1)<rp+rand()%5 )
			{
				move_dir = *dir;
			}
		}
		sprintf(move, "2 Move %c\n", move_dir);
		me.money-=2;
		return;
	}

	//evaluate the four directions
	move_dir='D';
	move_obstruction=0;
	i=BIG_NUMBER;  /* Unlikely to be chosen */
	total_near_robots=robot_proximity(me.x,me.y,1);
	me.angry += total_near_robots>8 ? 8 :total_near_robots;
	for (dir=s_nsew; *dir; dir++)
	{
		int obstruction=0;
		nx=me.x+DX(*dir); ny=me.y+DY(*dir);

		/* avoid some fights, maybe. */
		near_robots = 0;
		for (robot=robots; robot; robot=robot->next)
			if ( (robot->x==nx && robot->y==ny) )
		{
				flag_near = 1;
				obstruction+=((me.packages||me.angry>50) ? 8+rand()%16 : 6+rand()%9);
		}

#if 0
		// check for easy pickups
		if (0&&!obstruction)
		{
			if (me.max_load - me.load > 0)
				;
		}
#endif

		//look for danger/profit a move ahead here...? xxx

		near_robots=robot_proximity(nx,ny,1);
		if (near_robots) me.angry+=3;

		randomfactor  = rand()%5;

#		if 0  // trust in the path (mostly)
		if (near_robots>7)   
			randomfactor += (rand()%(near_robots/2));
#		endif

		if ( near_robots>2 && me.angry>20 && me.angry<2300 ) 
			randomfactor += (rand()%(me.angry>50?me.angry/10:5));

		if (lastmove==move_dir) randomfactor+=rand()%3;  //something different

		if (randomfactor>13 && me.angry<100)
			randomfactor=rand()%13;

		if (   (world[nx][ny].passable) 
		    && (i > goal->path->ds[nx][ny] + randomfactor + obstruction)  )
		{
			move_dir = *dir;
			move_obstruction = obstruction;
			i=goal->path->ds[nx][ny] + randomfactor + obstruction;
		}
	}

	nx=me.x+DX(move_dir); ny=me.y+DY(move_dir);
	for (robot=robots; robot; robot=robot->next)
		if (robot->x==nx && robot->y==ny)  
	{
		if (me.angry>14)
			bid += 10+(rand()%(me.angry/7)+1) + 2;
		else if (me.angry>5)
			bid += 8;
		else
			bid += move_obstruction/3;
	}

	if (me.angry>20 && total_near_robots)
	{
		bid += (total_near_robots+2)/3 * rand()%(me.angry/10);
	}
	else if (me.angry>5 && total_near_robots)
		bid++;

	if (!flag_near && me.angry>1)          me.angry = me.angry/2 + 1;
	if (!me.packages)                      bid /= 2;
	if (bid > me.money/10)                 bid = me.money/(flag_water?10:7);
	if (bid>1)                             bid = rand()%(bid);  /* be more random. */
	if (!me.packages && rand()%5==0)       bid = 1; /* no, even more random than that. */
	if (bid<=0)                            bid=1;
	if (move_dir=='D') printf("This path leads nowhere... I'm dead.\n");

	printf("Moving %c [%d] bound for (%d,%d)  \n", 
			move_dir, bid, goal->x, goal->y);
	if (bid>1)
		printf("bid %d.\n", bid);

	me.money-= (bid>0?bid:-bid);
	sprintf(move, "%d Move %c\n", bid, move_dir);
	lastmove=move_dir;
}

#define RUN_ERR(x) { errmsg=x; goto RUN_ERROR; }
char    move[8192];
void run()
{
	char *errmsg="No error.";
	char   *p;

	// don't need to remember goal ? maybe fast enough to 
	// eval every time....
	free_goal(&main_goal);     

	for (;;)
	{
		sprintf(move, "1 Drop\n");
		gmessage[0]='\0';  //xxx display msgs on console not implemented

		/* mark base as visited */
		if (world[me.x][me.y].type==BASE)
		{
			base_t *base;
			for (base=bases; base; base=base->next)
				if (base->x==me.x && base->y==me.y)
			{
				base->visited=turn;
				break;
			}
		}

		/* stay angry */
		if (me.angry<1) me.angry=1;       
		if (me.angry>500)
		{
			if (me.angry>10000)
				me.angry=rand()%5000;  /* but not too angry. */
			free_goal(&main_goal);     /* stop a moment. re-evaluate. */
		}
		me.angry+=robot_proximity(me.x, me.y, 1);

		main_goal = evaluate(main_goal);

		/* avoid water */
		if (water_threats(move))
			goto MOVE;

		/* drop any packages that need dropping */
		if (drop_packages(move))
			goto MOVE;

		/* pick up stuff */
		if (world[me.x][me.y].packages && do_pickups(move))
		{
			goto MOVE;
		}

		/* decide on move  -  you single-minded purposeful robot */

		do_move(move, main_goal);
		
MOVE:
		/* send move */
		szsend(sock, move);

		/* get result */
		if (0>=ss_read(sock, buffer, BUFSIZE-1)) RUN_ERR("read")
		if (update_status(buffer))               RUN_ERR("update game")

		/* list of packages {id x y weight} */
		p=buffer;
		if (0>nextline(sock, buffer, &p))        RUN_ERR("nextline")
		if (update_packages(buffer))             RUN_ERR("update packages")

	}
RUN_ERROR:
	fprintf(stderr, "Run error: %s\n", errmsg);
	printf("Dead, I guess.\n");
	shutdown(sock, 2);
	close(sock);
	return;
}
