/* 
 * Wrappers for malloc, free, etc. 
 */

#define ROUTER_MEMORY
#include "robot.h"

// Define critical sections



/*******************************************************************
**     int RouterMemoryConstruct (void)
**
**  Description:  Init the mutex.
*/
int RouterMemoryConstruct (void)
{
   // Initialize the critical section for stats management
#if 0
	mutex_init(&pAllocMutex);
   mutex_init(&MemoryCriticalSection);
#endif
	return 0;
}


/*
** int RouterMemoryDestruct (void)
**
** This code will free any resources used by this module.
**
** Returns SUCCESS automatically.
*/
int RouterMemoryDestruct (void)
{
#if 0
	mutex_destroy(&pAllocMutex);
   mutex_destroy(&MemoryCriticalSection);
#endif
	return 0;
}

int pMemoryInitFlag = 0;
typedef struct
{
	void *data;
	void *next;
	char *file;
	int  line;
	int  size;
} ALLOC;
ALLOC *Alist = NULL;

char *RouterStrdup(const char *s, char *file, int line)
{
 char *ret;
 ALLOC *A;
	if (ret = strdup(s))
	{
		A = malloc(sizeof(ALLOC));
		A->data = (void *)ret;
		A->size = strlen(s)+1;
		A->file = file;
		A->line = line;
		A->next = Alist;
		Alist = A;
	}
	return ret;
}

void *RouterCalloc(size_t n, size_t size, char *file, int line)
{
 void *ret;
 ALLOC *A;
	if (ret = calloc(n, size))
	{
		A = malloc(sizeof(ALLOC));
		A->data = ret;
		A->size = size * n;
		A->file = file;
		A->line = line;
		A->next = Alist;
		Alist = A;
	}
	return ret;
}

void *RouterMalloc(size_t size, char *file, int line)
{
 void *ret;
 ALLOC *A;
	if (ret = malloc(size))
	{
		A = malloc(sizeof(ALLOC));
		A->data = ret;
		A->size = size;
		A->file = file;
		A->line = line;
		A->next = Alist;
		Alist = A;
	}
	return ret;
}

void RouterFree(void *p, char *file, int line)
{
 ALLOC *A, *Aprev=NULL;
 	for (A = Alist; A; A=A->next)
	{
		if (A->data == p)
		{
			if (Aprev) Aprev->next = A->next;
			else       Alist = A->next;
			free(A);
			free(p);
			break;
		}
		Aprev = A;
	}
	//if (!A && !strstr(file, "ldap")) fprintf(stderr, "FREE WITH NO MALLOC! at %s, %d\n", file, line);
	if (!A) fprintf(stderr, "FREE WITH NO MALLOC! at %s, %d\n", file, line);
}

void *RouterRealloc(void *p, size_t size, char *file, int line)
{
 void *ret;
 ALLOC *A;
 	for (A = Alist; A; A=A->next)
		if (A->data == p)
			break;
	
	// do the realloc, record it if successful.
	ret = realloc(p, size);
	if (A)
	{
		A->data = ret;
		A->size = size;
	} 
	else if (p==NULL)
	{
		A = malloc(sizeof(ALLOC));
		A->data = ret;
		A->size = size;
		A->file = file;
		A->line = line;
		A->next = Alist;
		Alist = A;
	}

	if (!A) fprintf(stderr, "REALLOC WITH NO MALLOC! at %s, %d\n", file, line);

	return ret;
}

int memwrap(int cmd_printf())
{
	char *last_file=NULL;
	int last_line=0;
	int count=0;

	ALLOC *A;
 	for (A = Alist; A; A=A->next)
	{
		if (last_file != A->file || last_line != A->line)
		{
			if (count>0)
				cmd_printf("%s:%d   %d bytes. (x%d)\n", A->file, A->line, A->size, count+1);
			else
				cmd_printf("%s:%d   %d bytes.\n", A->file, A->line, A->size);
			count = 0;
		}
		else
			count++;
		last_file = A->file; last_line = A->line;
	}
	return 0;
}

int memwrap_clear(int cmd_printf())
{
	ALLOC *A, *Anext;
	 	for (A = Alist; A; A=Anext) { Anext=A->next; free(A); }
		Alist = NULL;
	cmd_printf("Allocation list cleared.\n");
	return 0;
}
