#ifdef ROBOT_MAIN
#define GLOBAL 
#else
#define GLOBAL extern
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <sys/types.h>

#include "ss.h"
#include "memwrap.h"
#define BUFSIZE 32768
#define MAXWORLD 1005

typedef struct
{
	int x;
	int y;
	int **ds;
	int n_robots;
} path_t;
path_t *find_path(int, int, int flag);

typedef struct 
{
	void *next;
	void *mnext;
	void *wnext;
	int n;
	int x;
	int y;
	void *owner;
	int weight;
	int dest_x;
	int dest_y;
	int pickup_flag;
	path_t *path;
	path_t *path_to_dest;
} package_t;

typedef struct 
{ 
	void *next;
	void *wnext;
	int n;
	int x;
	int y;
	int        money; 
	char      *name;
	int        max_load;
	int        load;
	package_t *packages;
	int        score;
	int        alive;
	int        angry;
} robot_t;

/* home base */
typedef struct
{
	void *next;
	int x; 
	int y;
	path_t *path;
	int visited;
} base_t;

typedef struct
{
	void *next;
	int x;
	int y; 
	enum { NONE, HOME, DELIVERY, PICKUP } type;
	path_t *path;
	int length; 
	int value;
	
} goal_t;

typedef struct
{
	enum { WALL, WATER, BASE, PLAIN } type;
	robot_t   *robot;
	package_t *packages;
	int value;
	int passable;
} world_t;

GLOBAL int             world_x, world_y;
GLOBAL world_t         world[MAXWORLD][MAXWORLD];
GLOBAL int             n_robots;
GLOBAL robot_t        *robots;
GLOBAL package_t      *packages;
GLOBAL base_t         *bases;
GLOBAL goal_t         *goal;

GLOBAL robot_t         me; 
GLOBAL int             sock;
GLOBAL char            buffer[BUFSIZE+2];
GLOBAL char            lastmove;
GLOBAL goal_t         *main_goal;
GLOBAL int             turn;
GLOBAL char            gmessage[1024];

/* Ah why not, might be useful: directions. */
#ifdef ROBOT_MAIN
const char s_nsew[5]="EWNS";
const int  i_nsew[4][2]={ {1, 0}, {-1,0}, {0, 1}, {0, -1} };
const int proximity[27][3] = 
{
	{ -2, -2, 1 } , { -1, -2, 2 } , {  0, -2, 3 } , {  1, -2, 2 } , {  2, -2, 1 } , 
	{ -2, -1, 2 } , { -1, -1, 5 } , {  0, -1, 8 } , {  1, -1, 5 } , {  2, -1, 2 } , 
	{ -2,  0, 3 } , { -1,  0, 8 } , {  0,  0,10 } , {  1,  0, 8 } , {  2,  0, 3 } , 
	{ -2,  1, 2 } , { -1,  1, 5 } , {  0,  1, 8 } , {  1,  1, 5 } , {  2,  1, 2 } , 
	{ -2,  2, 1 } , { -1,  2, 2 } , {  0,  2, 3 } , {  1,  2, 2 } , {  2,  2, 1 } , 
	{ 0, 0, 0 }
};
#else
GLOBAL int proximity[27][3];
GLOBAL char s_nsew[5];
GLOBAL int  i_nsew[4][2];
#endif

#define DX(c) (i_nsew[strchr(s_nsew, (c))-s_nsew][0])
#define DY(c) (i_nsew[strchr(s_nsew, (c))-s_nsew][1])
#define DIST(x,y) ( ((x)>(y) ? (x)-(y) : (y)-(x) ) )

#define PROX 26
#define PROX_DX 0
#define PROX_DY 1
#define PROX_WEIGHT  2

char *parseints(char *, int, ...);
void run(void);
int update_status(char *);
int update_packages(char *);
int nextline(int sock, char *buffer, char **p);
void display_board(void);

void free_goal(goal_t**);
void free_path(path_t**);
