#include "ss.h"
/*  Simple socket utils.  Very simple.  */

FILE *f;
struct protoent *pent;
char last_host[255];
unsigned int last_port;
struct sockaddr_in last_sin;

int ss_init(void)
{
#ifdef DEBUG
	if (!(f=fopen("tcp.log", "w")))
			return -1;
#endif
	last_port = 0;
	memset(last_host, 0x00, 254);
	memset(&last_sin, 0x00, sizeof(struct sockaddr_in));
	if (NULL==(pent = getprotobyname("tcp")) )
		return -1;

#ifdef __WINDOWS__
	{
    WORD wVersionRequested;             /*  We really want Winsock 1.1       */
    WSADATA wsaData;

    wVersionRequested = 0x0101;         /*  ... but we'll take 1.1           */
    if (WSAStartup (wVersionRequested, &wsaData) != 0)
        return -1;
	}
#endif

	return 0;
}


/* don't differentiate between connect errors on win32, since 
   I have no idea what the Windows error numbers are. */
#ifdef __WINDOWS__
#define ECONNREFUSED   -1
#define ETIMEDOUT      -2
#endif

int ss_open(const char *host, unsigned int port)
{
	struct sockaddr_in sin;
	struct hostent *phe;
	int s, rc;

	fprintf(stderr, "%s %d\n", host, port);

	if ( -1==(s = socket(AF_INET, SOCK_STREAM, pent->p_proto)) )
	{
		fprintf(stderr, "No socket.\n");
		last_port = 0;
		return -1;
	}

   /* the address lookup is expensive and not thread-safe anyway, so 
      if the previous host/port == current host/port, skip it.
    */
	if (0==last_host[0] || port!=last_port || strncmp(host, last_host, 254))
	{
		sin.sin_port = htons((short)port);
		sin.sin_family = AF_INET;

		/* check if it's a valid IP address */
		sin.sin_addr.s_addr = inet_addr(host);
		if (sin.sin_addr.s_addr == -1 )
		{
			if (phe = gethostbyname(host))
				memcpy(&sin.sin_addr, phe->h_addr, phe->h_length);
			else
			{
				shutdown(s, 2);
				close(s);
				last_port = 0;
				return -1;
			}
		}
		strncpy(last_host, host, 254);
		last_port = port;
		memcpy(&last_sin, &sin, sizeof(struct sockaddr_in));
	} else
	{
		memcpy(&sin, &last_sin, sizeof(struct sockaddr_in));
	}

	rc = connect(s, (struct sockaddr *)&sin, sizeof(struct sockaddr_in));
	if (rc)
	{
		switch (errno)
		{
			case ECONNREFUSED: return -1;
			case ETIMEDOUT:    return -1;
			default:           return -1;
		}
	} else 
		return s;
}

int ss_listen(unsigned int port)
{
	struct sockaddr_in sin;
	int s;
	int ov;

	s = socket(AF_INET, SOCK_STREAM, getprotobyname("tcp")->p_proto);
	if (!s)
	{
		fprintf(stderr, "No socket.\n");
		return 0;
	}

	ov = 1;
#ifndef __WINDOWS__
	setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &ov, sizeof(int));
#endif

	sin.sin_port = htons((short)port);
	sin.sin_family = AF_INET;

	/* check if it's a valid IP address */
	sin.sin_addr.s_addr = INADDR_ANY;
	if (bind(s, (struct sockaddr *)&sin, sizeof(sin)))
	{
		fprintf(stderr, "Could not bind.\n");
		return 0;
	}

	if (listen(s, 24))
	{
		fprintf(stderr, "Could not listen.\n");
		return 0;
	} else
		return s;
}

int ss_accept(int listener)
{
	struct sockaddr sa;
	int alen = sizeof(sa);

	return accept(listener, &sa, &alen);
}

int ss_read(int s, char *buf, size_t bufsiz)
{
	int rc;

	rc=read(s, buf, bufsiz);
	buf[ (rc > 0  ? rc : 0) ] = 0;
#ifdef DEBUG
	if (rc>0)
	{
		fprintf(f, " IN:%s", buf);
		fflush(f);
	}
#endif
	return rc;
}

int szsend(int s, const char *msg)
{
	int rc;
#ifdef DEBUG
	fprintf(f, "OUT:%s", msg);
	fflush(f);
#endif
	rc = send(s, msg, strlen(msg), 0);
	if (rc<0)
	{
		perror("Send error");
	}
	return rc;
}

